rfsrc.cart <- function(formula, data, ntree = 1, mtry = ncol(data), bootstrap = "none", nsplit = 0, ...)
{
  rfsrc(formula, data, ntree = ntree, mtry = mtry, bootstrap = bootstrap, nsplit = 0, ...)
}
