\name{rfsrc.news}
\alias{rfsrc.news}
\title{Show the NEWS file}
\description{
  Show the NEWS file of the \pkg{randomForestSRC} package.
}
\usage{
rfsrc.news(...)
}
\arguments{
  \item{...}{Further arguments passed to or from other methods.}
}
\value{
  None.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\keyword{documentation}
